@@ -15,9 +15,12 @@
 import collections
 import contextlib
 import functools
+import itertools
 import os
+import random
 import socket
 import ssl
+import sys
 import threading
 import time
 import uuid
@@ -29,6 +32,8 @@ import kombu.messaging
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import netutils
+from oslo_utils import versionutils
+import pkg_resources
 import six
 from six.moves.urllib import parse
 
@@ -36,6 +41,7 @@ from oslo_messaging._drivers import amqp as rpc_amqp
 from oslo_messaging._drivers import amqpdriver
 from oslo_messaging._drivers import base
 from oslo_messaging._drivers import common as rpc_common
+from oslo_messaging._drivers import pool
 from oslo_messaging._i18n import _
 from oslo_messaging._i18n import _LE
 from oslo_messaging._i18n import _LI
@@ -43,6 +49,9 @@ from oslo_messaging._i18n import _LW
 from oslo_messaging import _utils
 from oslo_messaging import exceptions
 
+# NOTE(sileht): don't exists in py2 socket module
+TCP_USER_TIMEOUT = 18
+
 
 rabbit_opts = [
     cfg.StrOpt('kombu_ssl_version',
@@ -71,27 +80,32 @@ rabbit_opts = [
                  deprecated_group='DEFAULT',
                  help='How long to wait before reconnecting in response to an '
                       'AMQP consumer cancel notification.'),
-    cfg.IntOpt('kombu_reconnect_timeout',
-               # NOTE(dhellmann): We want this to be similar to
-               # rpc_response_timeout, but we can't use
-               # "$rpc_response_timeout" as a default because that
-               # option may not have been defined by the time this
-               # option is accessed. Instead, document the intent in
-               # the help text for this option and provide a separate
-               # literal default value.
+    cfg.StrOpt('kombu_compression',
+               help="EXPERIMENTAL: Possible values are: gzip, bz2. If not "
+                    "set compression will not be used. This option may not"
+                    "be available in future versions."),
+    cfg.IntOpt('kombu_missing_consumer_retry_timeout',
+               deprecated_name="kombu_reconnect_timeout",
                default=60,
-               help='How long to wait before considering a reconnect '
-                    'attempt to have failed. This value should not be '
-                    'longer than rpc_response_timeout.'),
+               help='How long to wait a missing client beforce abandoning to '
+                    'send it its replies. This value should not be longer '
+                    'than rpc_response_timeout.'),
+    cfg.StrOpt('kombu_failover_strategy',
+               choices=('round-robin', 'shuffle'),
+               default='round-robin',
+               help='Determines how the next RabbitMQ node is chosen in case '
+                    'the one we are currently connected to becomes '
+                    'unavailable. Takes effect only if more than one '
+                    'RabbitMQ node is provided in config.'),
     cfg.StrOpt('rabbit_host',
                default='localhost',
                deprecated_group='DEFAULT',
                help='The RabbitMQ broker address where a single node is '
                     'used.'),
-    cfg.IntOpt('rabbit_port',
-               default=5672,
-               deprecated_group='DEFAULT',
-               help='The RabbitMQ broker port where a single node is used.'),
+    cfg.PortOpt('rabbit_port',
+                default=5672,
+                deprecated_group='DEFAULT',
+                help='The RabbitMQ broker port where a single node is used.'),
     cfg.ListOpt('rabbit_hosts',
                 default=['$rabbit_host:$rabbit_port'],
                 deprecated_group='DEFAULT',
@@ -125,6 +139,10 @@ rabbit_opts = [
                deprecated_group='DEFAULT',
                help='How long to backoff for between retries when connecting '
                     'to RabbitMQ.'),
+    cfg.IntOpt('rabbit_interval_max',
+               default=30,
+               help='Maximum interval of RabbitMQ connection retries. '
+                    'Default is 30 seconds.'),
     cfg.IntOpt('rabbit_max_retries',
                default=0,
                deprecated_group='DEFAULT',
@@ -133,9 +151,26 @@ rabbit_opts = [
     cfg.BoolOpt('rabbit_ha_queues',
                 default=False,
                 deprecated_group='DEFAULT',
-                help='Use HA queues in RabbitMQ (x-ha-policy: all). '
-                     'If you change this option, you must wipe the '
-                     'RabbitMQ database.'),
+                help='Try to use HA queues in RabbitMQ (x-ha-policy: all). '
+                'If you change this option, you must wipe the RabbitMQ '
+                'database. In RabbitMQ 3.0, queue mirroring is no longer '
+                'controlled by the x-ha-policy argument when declaring a '
+                'queue. If you just want to make sure that all queues (except '
+                ' those with auto-generated names) are mirrored across all '
+                'nodes, run: '
+                """\"rabbitmqctl set_policy HA '^(?!amq\.).*' """
+                """'{"ha-mode": "all"}' \""""),
+    cfg.IntOpt('rabbit_transient_queues_ttl',
+               min=1,
+               default=1800,
+               help='Positive integer representing duration in seconds for '
+                    'queue TTL (x-expires). Queues which are unused for the '
+                    'duration of the TTL are automatically deleted. The '
+                    'parameter affects only reply and fanout queues.'),
+    cfg.IntOpt('rabbit_qos_prefetch_count',
+               default=0,
+               help='Specifies the number of messages to prefetch. Setting to '
+                    'zero allows unlimited messages.'),
     cfg.IntOpt('heartbeat_timeout_threshold',
                default=60,
                help="Number of seconds after which the Rabbit broker is "
@@ -157,18 +192,39 @@ rabbit_opts = [
 LOG = logging.getLogger(__name__)
 
 
-def _get_queue_arguments(rabbit_ha_queues):
+def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl):
     """Construct the arguments for declaring a queue.
 
-    If the rabbit_ha_queues option is set, we declare a mirrored queue
+    If the rabbit_ha_queues option is set, we try to declare a mirrored queue
     as described here:
 
       http://www.rabbitmq.com/ha.html
 
     Setting x-ha-policy to all means that the queue will be mirrored
-    to all nodes in the cluster.
+    to all nodes in the cluster. In RabbitMQ 3.0, queue mirroring is
+    no longer controlled by the x-ha-policy argument when declaring a
+    queue. If you just want to make sure that all queues (except those
+    with auto-generated names) are mirrored across all nodes, run:
+      rabbitmqctl set_policy HA '^(?!amq\.).*' '{"ha-mode": "all"}'
+
+    If the rabbit_queue_ttl option is > 0, then the queue is
+    declared with the "Queue TTL" value as described here:
+
+      https://www.rabbitmq.com/ttl.html
+
+    Setting a queue TTL causes the queue to be automatically deleted
+    if it is unused for the TTL duration.  This is a helpful safeguard
+    to prevent queues with zero consumers from growing without bound.
     """
-    return {'x-ha-policy': 'all'} if rabbit_ha_queues else {}
+    args = {}
+
+    if rabbit_ha_queues:
+        args['x-ha-policy'] = 'all'
+
+    if rabbit_queue_ttl > 0:
+        args['x-expires'] = rabbit_queue_ttl * 1000
+
+    return args
 
 
 class RabbitMessage(dict):
@@ -191,26 +247,29 @@ class Consumer(object):
     """Consumer class."""
 
     def __init__(self, exchange_name, queue_name, routing_key, type, durable,
-                 auto_delete, callback, nowait=True, rabbit_ha_queues=None):
+                 exchange_auto_delete, queue_auto_delete, callback,
+                 nowait=True, rabbit_ha_queues=None, rabbit_queue_ttl=0):
         """Init the Publisher class with the exchange_name, routing_key,
         type, durable auto_delete
         """
         self.queue_name = queue_name
         self.exchange_name = exchange_name
         self.routing_key = routing_key
-        self.auto_delete = auto_delete
+        self.exchange_auto_delete = exchange_auto_delete
+        self.queue_auto_delete = queue_auto_delete
         self.durable = durable
         self.callback = callback
         self.type = type
         self.nowait = nowait
-        self.queue_arguments = _get_queue_arguments(rabbit_ha_queues)
+        self.queue_arguments = _get_queue_arguments(rabbit_ha_queues,
+                                                    rabbit_queue_ttl)
 
         self.queue = None
         self.exchange = kombu.entity.Exchange(
             name=exchange_name,
             type=type,
             durable=self.durable,
-            auto_delete=self.auto_delete)
+            auto_delete=self.exchange_auto_delete)
 
     def declare(self, conn):
         """Re-declare the queue after a rabbit (re)connect."""
@@ -219,7 +278,7 @@ class Consumer(object):
             channel=conn.channel,
             exchange=self.exchange,
             durable=self.durable,
-            auto_delete=self.auto_delete,
+            auto_delete=self.queue_auto_delete,
             routing_key=self.routing_key,
             queue_arguments=self.queue_arguments)
 
@@ -262,7 +321,6 @@ class Consumer(object):
         m2p = getattr(self.queue.channel, 'message_to_python', None)
         if m2p:
             message = m2p(message)
-
         try:
             self.callback(RabbitMessage(message))
         except Exception:
@@ -271,22 +329,10 @@ class Consumer(object):
             message.ack()
 
 
-class DummyConnectionLock(object):
-    def acquire(self):
-        pass
-
-    def release(self):
-        pass
-
+class DummyConnectionLock(_utils.DummyLock):
     def heartbeat_acquire(self):
         pass
 
-    def __enter__(self):
-        self.acquire()
-
-    def __exit__(self, type, value, traceback):
-        self.release()
-
 
 class ConnectionLock(DummyConnectionLock):
     """Lock object to protect access the the kombu connection
@@ -371,6 +417,7 @@ class Connection(object):
         self.max_retries = driver_conf.rabbit_max_retries
         self.interval_start = driver_conf.rabbit_retry_interval
         self.interval_stepping = driver_conf.rabbit_retry_backoff
+        self.interval_max = driver_conf.rabbit_interval_max
 
         self.login_method = driver_conf.rabbit_login_method
         self.fake_rabbit = driver_conf.fake_rabbit
@@ -380,6 +427,9 @@ class Connection(object):
         self.rabbit_userid = driver_conf.rabbit_userid
         self.rabbit_password = driver_conf.rabbit_password
         self.rabbit_ha_queues = driver_conf.rabbit_ha_queues
+        self.rabbit_transient_queues_ttl = \
+            driver_conf.rabbit_transient_queues_ttl
+        self.rabbit_qos_prefetch_count = driver_conf.rabbit_qos_prefetch_count
         self.heartbeat_timeout_threshold = \
             driver_conf.heartbeat_timeout_threshold
         self.heartbeat_rate = driver_conf.heartbeat_rate
@@ -387,7 +437,10 @@ class Connection(object):
         self.amqp_durable_queues = driver_conf.amqp_durable_queues
         self.amqp_auto_delete = driver_conf.amqp_auto_delete
         self.rabbit_use_ssl = driver_conf.rabbit_use_ssl
-        self.kombu_reconnect_timeout = driver_conf.kombu_reconnect_timeout
+        self.kombu_missing_consumer_retry_timeout = \
+            driver_conf.kombu_missing_consumer_retry_timeout
+        self.kombu_failover_strategy = driver_conf.kombu_failover_strategy
+        self.kombu_compression = driver_conf.kombu_compression
 
         if self.rabbit_use_ssl:
             self.kombu_ssl_version = driver_conf.kombu_ssl_version
@@ -399,9 +452,6 @@ class Connection(object):
         if self.max_retries <= 0:
             self.max_retries = None
 
-        # max retry-interval = 30 seconds
-        self.interval_max = 30
-
         if url.virtual_host is not None:
             virtual_host = url.virtual_host
         else:
@@ -409,15 +459,18 @@ class Connection(object):
 
         self._url = ''
         if self.fake_rabbit:
-            LOG.warn("Deprecated: fake_rabbit option is deprecated, set "
-                     "rpc_backend to kombu+memory or use the fake "
-                     "driver instead.")
+            LOG.warning(_LW("Deprecated: fake_rabbit option is deprecated, "
+                            "set rpc_backend to kombu+memory or use the fake "
+                            "driver instead."))
             self._url = 'memory://%s/' % virtual_host
         elif url.hosts:
             if url.transport.startswith('kombu+'):
-                LOG.warn(_LW('Selecting the kombu transport through the '
-                             'transport url (%s) is a experimental feature '
-                             'and this is not yet supported.') % url.transport)
+                LOG.warning(_LW('Selecting the kombu transport through the '
+                                'transport url (%s) is a experimental feature '
+                                'and this is not yet supported.'),
+                            url.transport)
+            if len(url.hosts) > 1:
+                random.shuffle(url.hosts)
             for host in url.hosts:
                 transport = url.transport.replace('kombu+', '')
                 transport = transport.replace('rabbit', 'amqp')
@@ -435,6 +488,8 @@ class Connection(object):
             transport = url.transport.replace('kombu+', '')
             self._url = "%s://%s" % (transport, virtual_host)
         else:
+            if len(self.rabbit_hosts) > 1:
+                random.shuffle(self.rabbit_hosts)
             for adr in self.rabbit_hosts:
                 hostname, port = netutils.parse_host_port(
                     adr, default_port=self.rabbit_port)
@@ -447,15 +502,19 @@ class Connection(object):
 
         self._initial_pid = os.getpid()
 
-        self._consumers = []
-        self._new_consumers = []
+        self._consumers = {}
+        self._new_tags = set()
+        self._active_tags = {}
+        self._tags = itertools.count(1)
+
         self._consume_loop_stopped = False
         self.channel = None
+        self.purpose = purpose
 
         # NOTE(sileht): if purpose is PURPOSE_LISTEN
         # we don't need the lock because we don't
         # have a heartbeat thread
-        if purpose == rpc_amqp.PURPOSE_SEND:
+        if purpose == rpc_common.PURPOSE_SEND:
             self._connection_lock = ConnectionLock()
         else:
             self._connection_lock = DummyConnectionLock()
@@ -463,23 +522,27 @@ class Connection(object):
         self.connection = kombu.connection.Connection(
             self._url, ssl=self._fetch_ssl_params(),
             login_method=self.login_method,
-            failover_strategy="shuffle",
             heartbeat=self.heartbeat_timeout_threshold,
+            failover_strategy=self.kombu_failover_strategy,
             transport_options={
                 'confirm_publish': True,
+                'client_properties': {'capabilities': {
+                    'authentication_failure_close': True,
+                    'connection.blocked': True,
+                    'consumer_cancel_notify': True}},
                 'on_blocked': self._on_connection_blocked,
                 'on_unblocked': self._on_connection_unblocked,
             },
         )
 
-        LOG.info(_LI('Connecting to AMQP server on %(hostname)s:%(port)s'),
-                 self.connection.info())
+        LOG.debug('Connecting to AMQP server on %(hostname)s:%(port)s',
+                  self.connection.info())
 
         # NOTE(sileht): kombu recommend to run heartbeat_check every
         # seconds, but we use a lock around the kombu connection
         # so, to not lock to much this lock to most of the time do nothing
         # expected waiting the events drain, we start heartbeat_check and
-        # retreive the server heartbeat packet only two times more than
+        # retrieve the server heartbeat packet only two times more than
         # the minimum required for the heartbeat works
         # (heatbeat_timeout/heartbeat_rate/2.0, default kombu
         # heartbeat_rate is 2)
@@ -495,13 +558,14 @@ class Connection(object):
         # the consume code does the heartbeat stuff
         # we don't need a thread
         self._heartbeat_thread = None
-        if purpose == rpc_amqp.PURPOSE_SEND:
+        if purpose == rpc_common.PURPOSE_SEND:
             self._heartbeat_start()
 
-        LOG.info(_LI('Connected to AMQP server on %(hostname)s:%(port)s'),
-                 self.connection.info())
+        LOG.debug('Connected to AMQP server on %(hostname)s:%(port)s '
+                  'via [%(transport)s] client',
+                  self.connection.info())
 
-        # NOTE(sileht): value choosen according the best practice from kombu
+        # NOTE(sileht): value chosen according the best practice from kombu
         # http://kombu.readthedocs.org/en/latest/reference/kombu.common.html#kombu.common.eventloop
         # For heatbeat, we can set a bigger timeout, and check we receive the
         # heartbeat packets regulary
@@ -583,7 +647,11 @@ class Connection(object):
         LOG.info(_LI("The broker has unblocked the connection"))
 
     def ensure_connection(self):
-        self.ensure(method=lambda: True)
+        # NOTE(sileht): we reset the channel and ensure
+        # the kombu underlying connection works
+        self._set_current_channel(None)
+        self.ensure(method=lambda: self.connection.connection)
+        self.set_transport_socket_timeout()
 
     def ensure(self, method, retry=None,
                recoverable_error_callback=None, error_callback=None,
@@ -599,10 +667,10 @@ class Connection(object):
 
         current_pid = os.getpid()
         if self._initial_pid != current_pid:
-            LOG.warn("Process forked after connection established! "
-                     "This can result in unpredictable behavior. "
-                     "See: http://docs.openstack.org/developer/"
-                     "oslo_messaging/transport.html")
+            LOG.warning(_LW("Process forked after connection established! "
+                            "This can result in unpredictable behavior. "
+                            "See: http://docs.openstack.org/developer/"
+                            "oslo.messaging/transport.html"))
             self._initial_pid = current_pid
 
         if retry is None:
@@ -651,12 +719,14 @@ class Connection(object):
             """Callback invoked when the kombu reconnects and creates
             a new channel, we use it the reconfigure our consumers.
             """
+
+            self.set_transport_socket_timeout()
             self._set_current_channel(new_channel)
             for consumer in self._consumers:
                 consumer.declare(self)
 
             LOG.info(_LI('Reconnected to AMQP server on '
-                         '%(hostname)s:%(port)s'),
+                         '%(hostname)s:%(port)s via [%(transport)s] client'),
                      self.connection.info())
 
         def execute_method(channel):
@@ -683,8 +753,8 @@ class Connection(object):
                 errback=on_error,
                 interval_start=self.interval_start or 1,
                 interval_step=self.interval_stepping,
-                on_revive=on_reconnection,
-            )
+                interval_max=self.interval_max,
+                on_revive=on_reconnection)
             ret, channel = autoretry_method()
             self._set_current_channel(channel)
             return ret
@@ -702,6 +772,10 @@ class Connection(object):
                     'tries: %(err_str)s') % info
             LOG.error(msg)
             raise exceptions.MessageDeliveryFailure(msg)
+        except rpc_amqp.AMQPDestinationNotFound:
+            # NOTE(sileht): we must reraise this without
+            # trigger error_callback
+            raise
         except Exception as exc:
             error_callback and error_callback(exc)
             raise
@@ -711,11 +785,26 @@ class Connection(object):
 
         NOTE(sileht): Must be called within the connection lock
         """
-        if self.channel is not None and new_channel != self.channel:
+        if new_channel == self.channel:
+            return
+
+        if self.channel is not None:
             self.PUBLISHER_DECLARED_QUEUES.pop(self.channel, None)
             self.connection.maybe_close_channel(self.channel)
+
         self.channel = new_channel
 
+        if (new_channel is not None and
+           self.purpose == rpc_common.PURPOSE_LISTEN):
+            self._set_qos(new_channel)
+
+    def _set_qos(self, channel):
+        """Set QoS prefetch count on the channel"""
+        if self.rabbit_qos_prefetch_count > 0:
+            channel.basic_qos(0,
+                              self.rabbit_qos_prefetch_count,
+                              False)
+
     def close(self):
         """Close/release this connection."""
         self._heartbeat_stop()
@@ -731,12 +820,14 @@ class Connection(object):
 
         with self._connection_lock:
             try:
-                for tag, consumer in enumerate(self._consumers):
+                for consumer, tag in self._consumers.items():
                     consumer.cancel(tag=tag)
             except recoverable_errors:
-                self._set_current_channel(None)
                 self.ensure_connection()
-            self._consumers = []
+            self._consumers.clear()
+            self._active_tags.clear()
+            self._new_tags.clear()
+            self._tags = itertools.count(1)
 
     def _heartbeat_supported_and_enabled(self):
         if self.heartbeat_timeout_threshold <= 0:
@@ -745,32 +836,53 @@ class Connection(object):
         if self.connection.supports_heartbeats:
             return True
         elif not self._heartbeat_support_log_emitted:
-            LOG.warn(_LW("Heartbeat support requested but it is not supported "
-                         "by the kombu driver or the broker"))
+            LOG.warning(_LW("Heartbeat support requested but it is not "
+                            "supported by the kombu driver or the broker"))
             self._heartbeat_support_log_emitted = True
         return False
 
-    @contextlib.contextmanager
-    def _transport_socket_timeout(self, timeout):
+    def set_transport_socket_timeout(self, timeout=None):
         # NOTE(sileht): they are some case where the heartbeat check
         # or the producer.send return only when the system socket
         # timeout if reach. kombu doesn't allow use to customise this
         # timeout so for py-amqp we tweak ourself
-        sock = getattr(self.connection.transport, 'sock', None)
-        if sock:
-            orig_timeout = sock.gettimeout()
+        # NOTE(dmitryme): Current approach works with amqp==1.4.9 and
+        # kombu==3.0.33. Once the commit below is released, we should
+        # try to set the socket timeout in the constructor:
+        # https://github.com/celery/py-amqp/pull/64
+
+        heartbeat_timeout = self.heartbeat_timeout_threshold
+        if self._heartbeat_supported_and_enabled():
+            # NOTE(sileht): we are supposed to send heartbeat every
+            # heartbeat_timeout, no need to wait more otherwise will
+            # disconnect us, so raise timeout earlier ourself
+            if timeout is None:
+                timeout = heartbeat_timeout
+            else:
+                timeout = min(heartbeat_timeout, timeout)
+
+        try:
+            sock = self.channel.connection.sock
+        except AttributeError as e:
+            # Level is set to debug because otherwise we would spam the logs
+            LOG.debug('Failed to get socket attribute: %s' % str(e))
+        else:
             sock.settimeout(timeout)
+            if sys.platform != 'win32':
+                sock.setsockopt(socket.IPPROTO_TCP,
+                                TCP_USER_TIMEOUT,
+                                timeout * 1000 if timeout is not None else 0)
+
+    @contextlib.contextmanager
+    def _transport_socket_timeout(self, timeout):
+        self.set_transport_socket_timeout(timeout)
         yield
-        if sock:
-            sock.settimeout(orig_timeout)
+        self.set_transport_socket_timeout()
 
     def _heartbeat_check(self):
-        # NOTE(sileht): we are suposed to send at least one heartbeat
+        # NOTE(sileht): we are supposed to send at least one heartbeat
         # every heartbeat_timeout_threshold, so no need to way more
-        with self._transport_socket_timeout(
-                self.heartbeat_timeout_threshold):
-            self.connection.heartbeat_check(
-                rate=self.heartbeat_rate)
+        self.connection.heartbeat_check(rate=self.heartbeat_rate)
 
     def _heartbeat_start(self):
         if self._heartbeat_supported_and_enabled():
@@ -836,8 +948,13 @@ class Connection(object):
 
         def _declare_consumer():
             consumer.declare(self)
-            self._consumers.append(consumer)
-            self._new_consumers.append(consumer)
+            tag = self._active_tags.get(consumer.queue_name)
+            if tag is None:
+                tag = next(self._tags)
+                self._active_tags[consumer.queue_name] = tag
+                self._new_tags.add(tag)
+
+            self._consumers[consumer] = tag
             return consumer
 
         with self._connection_lock:
@@ -855,7 +972,8 @@ class Connection(object):
             raise rpc_common.Timeout()
 
         def _recoverable_error_callback(exc):
-            self._new_consumers = self._consumers
+            if not isinstance(exc, rpc_common.Timeout):
+                self._new_tags = set(self._consumers.values())
             timer.check_return(_raise_timeout, exc)
 
         def _error_callback(exc):
@@ -864,18 +982,18 @@ class Connection(object):
                       exc)
 
         def _consume():
-            # NOTE(sileht): in case the acknowledgement or requeue of a
+            # NOTE(sileht): in case the acknowledgment or requeue of a
             # message fail, the kombu transport can be disconnected
             # In this case, we must redeclare our consumers, so raise
             # a recoverable error to trigger the reconnection code.
             if not self.connection.connected:
                 raise self.connection.recoverable_connection_errors[0]
 
-            if self._new_consumers:
-                for tag, consumer in enumerate(self._consumers):
-                    if consumer in self._new_consumers:
+            if self._new_tags:
+                for consumer, tag in self._consumers.items():
+                    if tag in self._new_tags:
                         consumer.consume(tag=tag)
-                self._new_consumers = []
+                        self._new_tags.remove(tag)
 
             poll_timeout = (self._poll_timeout if timeout is None
                             else min(timeout, self._poll_timeout))
@@ -912,9 +1030,11 @@ class Connection(object):
                             routing_key=topic,
                             type='direct',
                             durable=False,
-                            auto_delete=True,
+                            exchange_auto_delete=True,
+                            queue_auto_delete=False,
                             callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues)
+                            rabbit_ha_queues=self.rabbit_ha_queues,
+                            rabbit_queue_ttl=self.rabbit_transient_queues_ttl)
 
         self.declare_consumer(consumer)
 
@@ -926,7 +1046,8 @@ class Connection(object):
                             routing_key=topic,
                             type='topic',
                             durable=self.amqp_durable_queues,
-                            auto_delete=self.amqp_auto_delete,
+                            exchange_auto_delete=self.amqp_auto_delete,
+                            queue_auto_delete=self.amqp_auto_delete,
                             callback=callback,
                             rabbit_ha_queues=self.rabbit_ha_queues)
 
@@ -944,9 +1065,11 @@ class Connection(object):
                             routing_key=topic,
                             type='fanout',
                             durable=False,
-                            auto_delete=True,
+                            exchange_auto_delete=True,
+                            queue_auto_delete=False,
                             callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues)
+                            rabbit_ha_queues=self.rabbit_ha_queues,
+                            rabbit_queue_ttl=self.rabbit_transient_queues_ttl)
 
         self.declare_consumer(consumer)
 
@@ -965,38 +1088,36 @@ class Connection(object):
         with self._connection_lock:
             self.ensure(method, retry=retry, error_callback=_error_callback)
 
+    def _get_expiration(self, timeout):
+        # NOTE(gcb) kombu accept TTL as seconds instead of millisecond since
+        # version 3.0.25, so do conversion according to kombu version.
+        # TODO(gcb) remove this workaround when all supported branches
+        # with requirement kombu >=3.0.25
+        if timeout is not None:
+            kombu_version = pkg_resources.get_distribution('kombu').version
+            if not versionutils.is_compatible('3.0.25', kombu_version):
+                timeout = int(timeout * 1000)
+
+        return timeout
+
     def _publish(self, exchange, msg, routing_key=None, timeout=None):
         """Publish a message."""
         producer = kombu.messaging.Producer(exchange=exchange,
                                             channel=self.channel,
+                                            auto_declare=not exchange.passive,
                                             routing_key=routing_key)
 
-        expiration = None
-        if timeout:
-            # AMQP TTL is in milliseconds when set in the property.
-            # Details: http://www.rabbitmq.com/ttl.html#per-message-ttl
-            expiration = int(timeout * 1000)
-
-        # NOTE(sileht): no need to wait more, caller expects
-        # a answer before timeout is reached
-        transport_timeout = timeout
-
-        heartbeat_timeout = self.heartbeat_timeout_threshold
-        if (self._heartbeat_supported_and_enabled() and (
-                transport_timeout is None or
-                transport_timeout > heartbeat_timeout)):
-            # NOTE(sileht): we are supposed to send heartbeat every
-            # heartbeat_timeout, no need to wait more otherwise will
-            # disconnect us, so raise timeout earlier ourself
-            transport_timeout = heartbeat_timeout
-
         log_info = {'msg': msg,
                     'who': exchange or 'default',
                     'key': routing_key}
         LOG.trace('Connection._publish: sending message %(msg)s to'
                   ' %(who)s with routing key %(key)s', log_info)
-        with self._transport_socket_timeout(transport_timeout):
-            producer.publish(msg, expiration=expiration)
+
+        # NOTE(sileht): no need to wait more, caller expects
+        # a answer before timeout is reached
+        with self._transport_socket_timeout(timeout):
+            producer.publish(msg, expiration=self._get_expiration(timeout),
+                             compression=self.kombu_compression)
 
     # List of notification queue declared on the channel to avoid
     # unnecessary redeclaration. This list is resetted each time
@@ -1030,7 +1151,7 @@ class Connection(object):
                 auto_delete=exchange.auto_delete,
                 name=routing_key,
                 routing_key=routing_key,
-                queue_arguments=_get_queue_arguments(self.rabbit_ha_queues))
+                queue_arguments=_get_queue_arguments(self.rabbit_ha_queues, 0))
             log_info = {'key': routing_key, 'exchange': exchange}
             LOG.trace(
                 'Connection._publish_and_creates_default_queue: '
@@ -1040,32 +1161,20 @@ class Connection(object):
 
         self._publish(exchange, msg, routing_key=routing_key, timeout=timeout)
 
-    def _publish_and_retry_on_missing_exchange(self, exchange, msg,
-                                               routing_key=None, timeout=None):
-        """Publisher that retry if the exchange is missing.
-        """
-
+    def _publish_and_raises_on_missing_exchange(self, exchange, msg,
+                                                routing_key=None,
+                                                timeout=None):
+        """Publisher that raises exception if exchange is missing."""
         if not exchange.passive:
             raise RuntimeError("_publish_and_retry_on_missing_exchange() must "
                                "be called with an passive exchange.")
 
-        # TODO(sileht): use @retrying
-        # NOTE(sileht): no need to wait the application expect a response
-        # before timeout is exshauted
-        duration = (
-            timeout if timeout is not None
-            else self.kombu_reconnect_timeout
-        )
-
-        timer = rpc_common.DecayingTimer(duration=duration)
-        timer.start()
-
-        while True:
-            try:
-                self._publish(exchange, msg, routing_key=routing_key,
-                              timeout=timeout)
-                return
-            except self.connection.channel_errors as exc:
+        try:
+            self._publish(exchange, msg, routing_key=routing_key,
+                          timeout=timeout)
+            return
+        except self.connection.channel_errors as exc:
+            if exc.code == 404:
                 # NOTE(noelbk/sileht):
                 # If rabbit dies, the consumer can be disconnected before the
                 # publisher sends, and if the consumer hasn't declared the
@@ -1074,24 +1183,9 @@ class Connection(object):
                 # So we set passive=True to the publisher exchange and catch
                 # the 404 kombu ChannelError and retry until the exchange
                 # appears
-                if exc.code == 404 and timer.check_return() > 0:
-                    LOG.info(_LI("The exchange %(exchange)s to send to "
-                                 "%(routing_key)s doesn't exist yet, "
-                                 "retrying...") % {
-                                     'exchange': exchange.name,
-                                     'routing_key': routing_key})
-                    time.sleep(0.25)
-                    continue
-                elif exc.code == 404:
-                    msg = _("The exchange %(exchange)s to send to "
-                            "%(routing_key)s still doesn't exist after "
-                            "%(duration)s sec abandonning...") % {
-                                'duration': duration,
-                                'exchange': exchange.name,
-                                'routing_key': routing_key}
-                    LOG.info(msg)
-                    raise rpc_amqp.AMQPDestinationNotFound(msg)
-                raise
+                raise rpc_amqp.AMQPDestinationNotFound(
+                    "exchange %s doesn't exists" % exchange.name)
+            raise
 
     def direct_send(self, msg_id, msg):
         """Send a 'direct' message."""
@@ -1101,7 +1195,7 @@ class Connection(object):
                                          auto_delete=True,
                                          passive=True)
 
-        self._ensure_publishing(self._publish_and_retry_on_missing_exchange,
+        self._ensure_publishing(self._publish_and_raises_on_missing_exchange,
                                 exchange, msg, routing_key=msg_id)
 
     def topic_send(self, exchange_name, topic, msg, timeout=None, retry=None):
@@ -1113,7 +1207,8 @@ class Connection(object):
             auto_delete=self.amqp_auto_delete)
 
         self._ensure_publishing(self._publish, exchange, msg,
-                                routing_key=topic, retry=retry)
+                                routing_key=topic, timeout=timeout,
+                                retry=retry)
 
     def fanout_send(self, topic, msg, retry=None):
         """Send a 'fanout' message."""
@@ -1157,7 +1252,13 @@ class RabbitDriver(amqpdriver.AMQPDriverBase):
         conf.register_opts(rpc_amqp.amqp_opts, group=opt_group)
         conf.register_opts(base.base_opts, group=opt_group)
 
-        connection_pool = rpc_amqp.ConnectionPool(
+        self.missing_destination_retry_timeout = (
+            conf.oslo_messaging_rabbit.kombu_missing_consumer_retry_timeout)
+
+        self.prefetch_size = (
+            conf.oslo_messaging_rabbit.rabbit_qos_prefetch_count)
+
+        connection_pool = pool.ConnectionPool(
             conf, conf.oslo_messaging_rabbit.rpc_conn_pool_size,
             url, Connection)
 
@@ -1165,8 +1266,7 @@ class RabbitDriver(amqpdriver.AMQPDriverBase):
             conf, url,
             connection_pool,
             default_exchange,
-            allowed_remote_exmods,
-            conf.oslo_messaging_rabbit.send_single_reply,
+            allowed_remote_exmods
         )
 
     def require_features(self, requeue=True):
