@@ -36,6 +36,7 @@ from six import moves
 
 from oslo_messaging._drivers.protocols.amqp import eventloop
 from oslo_messaging._drivers.protocols.amqp import opts
+from oslo_messaging._i18n import _LE, _LI, _LW
 from oslo_messaging import exceptions
 from oslo_messaging import transport
 
@@ -72,6 +73,10 @@ class Replies(pyngus.ReceiverEventHandler):
         self._credit = 0
         self._receiver.open()
 
+    def destroy(self):
+        self._correlation = None
+        self._receiver = None
+
     def ready(self):
         return self._ready
 
@@ -86,8 +91,8 @@ class Replies(pyngus.ReceiverEventHandler):
         # reply is placed on reply_queue
         self._correlation[request.id] = reply_queue
         request.reply_to = self._receiver.source_address
-        LOG.debug("Reply for msg id=%s expected on link %s",
-                  request.id, request.reply_to)
+        LOG.debug("Reply for msg id=%(id)s expected on link %(reply_to)s",
+                  {'id': request.id, 'reply_to': request.reply_to})
         return request.id
 
     def cancel_response(self, msg_id):
@@ -117,7 +122,7 @@ class Replies(pyngus.ReceiverEventHandler):
         # TODO(kgiusti) Unclear if this error will ever occur (as opposed to
         # the Connection failing instead).  Log for now, possibly implement a
         # recovery strategy if necessary.
-        LOG.error("Reply subscription closed by peer: %s",
+        LOG.error(_LE("Reply subscription closed by peer: %s"),
                   (pn_condition or "no error given"))
 
     def message_received(self, receiver, message, handle):
@@ -137,8 +142,8 @@ class Replies(pyngus.ReceiverEventHandler):
             del self._correlation[key]
             receiver.message_accepted(handle)
         else:
-            LOG.warn("Can't find receiver for response msg id=%s, dropping!",
-                     key)
+            LOG.warning(_LW("Can't find receiver for response msg id=%s, "
+                            "dropping!"), key)
             receiver.message_modified(handle, True, True, None)
 
     def _update_credit(self):
@@ -157,6 +162,7 @@ class Server(pyngus.ReceiverEventHandler):
         self._incoming = incoming
         self._addresses = addresses
         self._capacity = 500   # credit per link
+        self._receivers = None
 
     def attach(self, connection):
         """Create receiver links over the given connection for all the
@@ -180,18 +186,21 @@ class Server(pyngus.ReceiverEventHandler):
             r.open()
             self._receivers.append(r)
 
+    def destroy(self):
+        self._receivers = None
+
     # Pyngus ReceiverLink event callbacks:
 
     def receiver_remote_closed(self, receiver, pn_condition):
         """This is a Pyngus callback, invoked by Pyngus when the peer of this
         receiver link has initiated closing the connection.
         """
-        text = "Server subscription %(addr)s closed by peer: %(err_msg)s"
         vals = {
             "addr": receiver.source_address or receiver.target_address,
             "err_msg": pn_condition or "no error given"
         }
-        LOG.error(text % vals)
+        LOG.error(_LE("Server subscription %(addr)s closed "
+                      "by peer: %(err_msg)s"), vals)
 
     def message_received(self, receiver, message, handle):
         """This is a Pyngus callback, invoked by Pyngus when a new message
@@ -206,23 +215,26 @@ class Server(pyngus.ReceiverEventHandler):
 
 class Hosts(object):
     """An order list of TransportHost addresses. Connection failover
-    progresses from one host to the next.
+    progresses from one host to the next.  username and password come from the
+    configuration and are used only if no username/password was given in the
+    URL.
     """
-    def __init__(self, entries=None):
-        self._entries = entries[:] if entries else []
+    def __init__(self, entries=None, default_username=None,
+                 default_password=None):
+        if entries:
+            self._entries = entries[:]
+        else:
+            self._entries = [transport.TransportHost(hostname="localhost",
+                                                     port=5672)]
         for entry in self._entries:
             entry.port = entry.port or 5672
+            entry.username = entry.username or default_username
+            entry.password = entry.password or default_password
         self._current = 0
 
-    def add(self, transport_host):
-        self._entries.append(transport_host)
-
     @property
     def current(self):
-        if len(self._entries):
-            return self._entries[self._current]
-        else:
-            return transport.TransportHost(hostname="localhost", port=5672)
+        return self._entries[self._current]
 
     def next(self):
         if len(self._entries) > 1:
@@ -245,6 +257,7 @@ class Controller(pyngus.ConnectionEventHandler):
     """
     def __init__(self, hosts, default_exchange, config):
         self.processor = None
+        self._socket_connection = None
         # queue of Task() objects to execute on the eventloop once the
         # connection is ready:
         self._tasks = moves.queue.Queue(maxsize=500)
@@ -256,7 +269,6 @@ class Controller(pyngus.ConnectionEventHandler):
         self._senders = {}
         # Servers (set of receiving links), indexed by target:
         self._servers = {}
-        self.hosts = Hosts(hosts)
 
         opt_group = cfg.OptGroup(name='oslo_messaging_amqp',
                                  title='AMQP 1.0 driver options')
@@ -277,6 +289,11 @@ class Controller(pyngus.ConnectionEventHandler):
         self.ssl_key_password = config.oslo_messaging_amqp.ssl_key_password
         self.ssl_allow_insecure = \
             config.oslo_messaging_amqp.allow_insecure_clients
+        self.sasl_mechanisms = config.oslo_messaging_amqp.sasl_mechanisms
+        self.sasl_config_dir = config.oslo_messaging_amqp.sasl_config_dir
+        self.sasl_config_name = config.oslo_messaging_amqp.sasl_config_name
+        self.hosts = Hosts(hosts, config.oslo_messaging_amqp.username,
+                           config.oslo_messaging_amqp.password)
         self.separator = "."
         self.fanout_qualifier = "all"
         self.default_exchange = default_exchange
@@ -310,6 +327,15 @@ class Controller(pyngus.ConnectionEventHandler):
             LOG.debug("Waiting for eventloop to exit")
             self.processor.shutdown(wait, timeout)
             self.processor = None
+        self._tasks = None
+        self._senders = None
+        for server in self._servers.values():
+            server.destroy()
+        self._servers.clear()
+        self._socket_connection = None
+        if self._replies:
+            self._replies.destroy()
+            self._replies = None
         LOG.debug("Eventloop exited, driver shut down")
 
     # The remaining methods are reserved to run from the eventloop thread only!
@@ -323,7 +349,8 @@ class Controller(pyngus.ConnectionEventHandler):
         will include the reply message (if successful).
         """
         address = self._resolve(target)
-        LOG.debug("Sending request for %s to %s", target, address)
+        LOG.debug("Sending request for %(target)s to %(address)s",
+                  {'target': target, 'address': address})
         if reply_expected:
             msg_id = self._replies.prepare_for_response(request, result_queue)
 
@@ -374,7 +401,8 @@ class Controller(pyngus.ConnectionEventHandler):
         self._subscribe(target, addresses, in_queue)
 
     def _subscribe(self, target, addresses, in_queue):
-        LOG.debug("Subscribing to %s (%s)", target, addresses)
+        LOG.debug("Subscribing to %(target)s (%(addresses)s)",
+                  {'target': target, 'addresses': addresses})
         self._servers[target] = Server(addresses, in_queue)
         self._servers[target].attach(self._socket_connection.connection)
 
@@ -435,7 +463,7 @@ class Controller(pyngus.ConnectionEventHandler):
     def _do_connect(self):
         """Establish connection and reply subscription on processor thread."""
         host = self.hosts.current
-        conn_props = {}
+        conn_props = {'hostname': host.hostname}
         if self.idle_timeout:
             conn_props["idle-time-out"] = float(self.idle_timeout)
         if self.trace_protocol:
@@ -451,6 +479,14 @@ class Controller(pyngus.ConnectionEventHandler):
                                             self.ssl_key_file,
                                             self.ssl_key_password)
             conn_props["x-ssl-allow-cleartext"] = self.ssl_allow_insecure
+        # SASL configuration:
+        if self.sasl_mechanisms:
+            conn_props["x-sasl-mechs"] = self.sasl_mechanisms
+        if self.sasl_config_dir:
+            conn_props["x-sasl-config-dir"] = self.sasl_config_dir
+        if self.sasl_config_name:
+            conn_props["x-sasl-config-name"] = self.sasl_config_name
+
         self._socket_connection = self.processor.connect(host,
                                                          handler=self,
                                                          properties=conn_props)
@@ -467,7 +503,7 @@ class Controller(pyngus.ConnectionEventHandler):
             try:
                 self._tasks.get(False).execute(self)
             except Exception as e:
-                LOG.exception("Error processing task: %s", e)
+                LOG.exception(_LE("Error processing task: %s"), e)
             count += 1
 
         # if we hit _max_task_batch, resume task processing later:
@@ -499,7 +535,7 @@ class Controller(pyngus.ConnectionEventHandler):
         """Called when the driver destroys the controller, this method attempts
         to cleanly close the AMQP connection to the peer.
         """
-        LOG.info("Shutting down AMQP connection")
+        LOG.info(_LI("Shutting down AMQP connection"))
         self._closing = True
         if self._socket_connection.connection.active:
             # try a clean shutdown
@@ -514,8 +550,9 @@ class Controller(pyngus.ConnectionEventHandler):
         """Invoked when the Replies reply link has become active.  At this
         point, we are ready to send/receive messages (via Task processing).
         """
-        LOG.info("Messaging is active (%s:%i)", self.hosts.current.hostname,
-                 self.hosts.current.port)
+        LOG.info(_LI("Messaging is active (%(hostname)s:%(port)s)"),
+                 {'hostname': self.hosts.current.hostname,
+                  'port': self.hosts.current.port})
         self._schedule_task_processing()
 
     # callback from eventloop on socket error
@@ -543,9 +580,10 @@ class Controller(pyngus.ConnectionEventHandler):
         the peer is up.  At this point, the driver will activate all subscriber
         links (server) and the reply link.
         """
-        LOG.debug("Connection active (%s:%i), subscribing...",
-                  self.hosts.current.hostname, self.hosts.current.port)
-        for s in self._servers.itervalues():
+        LOG.debug("Connection active (%(hostname)s:%(port)s), subscribing...",
+                  {'hostname': self.hosts.current.hostname,
+                   'port': self.hosts.current.port})
+        for s in self._servers.values():
             s.attach(self._socket_connection.connection)
         self._replies = Replies(self._socket_connection.connection,
                                 lambda: self._reply_link_ready())
@@ -570,7 +608,7 @@ class Controller(pyngus.ConnectionEventHandler):
             # connection. Acknowledge the close, and try to reconnect/failover
             # later once the connection has closed (connection_closed is
             # called).
-            LOG.info("Connection closed by peer: %s",
+            LOG.info(_LI("Connection closed by peer: %s"),
                      reason or "no reason given")
             self._socket_connection.connection.close()
 
@@ -581,9 +619,11 @@ class Controller(pyngus.ConnectionEventHandler):
         """
         if outcome == proton.SASL.OK:
             return
-        LOG.error("AUTHENTICATION FAILURE: Cannot connect to %s:%s as user %s",
-                  self.hosts.current.hostname, self.hosts.current.port,
-                  self.hosts.current.username)
+        LOG.error(_LE("AUTHENTICATION FAILURE: Cannot connect to "
+                      "%(hostname)s:%(port)s as user %(username)s"),
+                  {'hostname': self.hosts.current.hostname,
+                   'port': self.hosts.current.port,
+                   'username': self.hosts.current.username})
         # connection failure will be handled later
 
     def _complete_shutdown(self):
@@ -592,7 +632,7 @@ class Controller(pyngus.ConnectionEventHandler):
         """
         self._socket_connection.close()
         self.processor.shutdown()
-        LOG.info("Messaging has shutdown")
+        LOG.info(_LI("Messaging has shutdown"))
 
     def _handle_connection_loss(self):
         """The connection to the messaging service has been lost.  Try to
@@ -608,10 +648,12 @@ class Controller(pyngus.ConnectionEventHandler):
             if not self._reconnecting:
                 self._reconnecting = True
                 self._replies = None
-                d = self._delay
-                LOG.info("delaying reconnect attempt for %d seconds", d)
-                self.processor.schedule(lambda: self._do_reconnect(), d)
-                self._delay = 1 if self._delay == 0 else min(d * 2, 60)
+                LOG.info(_LI("delaying reconnect attempt for %d seconds"),
+                         self._delay)
+                self.processor.schedule(lambda: self._do_reconnect(),
+                                        self._delay)
+                self._delay = (1 if self._delay == 0
+                               else min(self._delay * 2, 60))
 
     def _do_reconnect(self):
         """Invoked on connection/socket failure, failover and re-connect to the
@@ -625,5 +667,6 @@ class Controller(pyngus.ConnectionEventHandler):
             self._senders = {}
             self._socket_connection.reset()
             host = self.hosts.next()
-            LOG.info("Reconnecting to: %s:%i", host.hostname, host.port)
+            LOG.info(_LI("Reconnecting to: %(hostname)s:%(port)s"),
+                     {'hostname': host.hostname, 'port': host.port})
             self._socket_connection.connect(host)
