@@ -13,6 +13,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import eventlet
 import threading
 
 from oslo_config import cfg
@@ -20,30 +21,46 @@ import testscenarios
 
 import mock
 import oslo_messaging
+from oslo_messaging import server as server_module
 from oslo_messaging.tests import utils as test_utils
-from six.moves import mock
 
 load_tests = testscenarios.load_tests_apply_scenarios
 
 
 class ServerSetupMixin(object):
 
-    class Server(object):
+    class Server(threading.Thread):
         def __init__(self, transport, topic, server, endpoint, serializer):
+            self.controller = ServerSetupMixin.ServerController()
             target = oslo_messaging.Target(topic=topic, server=server)
-            self._server = oslo_messaging.get_rpc_server(transport,
-                                                         target,
-                                                         [endpoint, self],
-                                                         serializer=serializer)
+            self.server = oslo_messaging.get_rpc_server(transport,
+                                                        target,
+                                                        [endpoint,
+                                                         self.controller],
+                                                        serializer=serializer)
+
+            super(ServerSetupMixin.Server, self).__init__()
+            self.daemon = True
+
+        def wait(self):
+            # Wait for the executor to process the stop message, indicating all
+            # test messages have been processed
+            self.controller.stopped.wait()
+
+            # Check start() does nothing with a running server
+            self.server.start()
+            self.server.stop()
+            self.server.wait()
+
+        def run(self):
+            self.server.start()
+
+    class ServerController(object):
+        def __init__(self):
+            self.stopped = threading.Event()
 
         def stop(self, ctxt):
-            # Check start() does nothing with a running server
-            self._server.start()
-            self._server.stop()
-            self._server.wait()
-
-        def start(self):
-            self._server.start()
+            self.stopped.set()
 
     class TestSerializer(object):
 
@@ -73,13 +90,13 @@ class ServerSetupMixin(object):
         thread.daemon = True
         thread.start()
 
-        return thread
+        return server
 
-    def _stop_server(self, client, server_thread, topic=None):
+    def _stop_server(self, client, server, topic=None):
         if topic is not None:
             client = client.prepare(topic=topic)
         client.cast({}, 'stop')
-        server_thread.join(timeout=30)
+        server.wait()
 
     def _setup_client(self, transport, topic='testtopic'):
         return oslo_messaging.RPCClient(transport,
@@ -110,7 +127,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self.assertIsInstance(server.dispatcher, oslo_messaging.RPCDispatcher)
         self.assertIs(server.dispatcher.endpoints, endpoints)
         self.assertIs(server.dispatcher.serializer, serializer)
-        self.assertEqual('blocking', server.executor)
+        self.assertEqual('blocking', server.executor_type)
 
     def test_server_wait_method(self):
         transport = oslo_messaging.get_transport(self.conf, url='fake:')
@@ -118,64 +135,32 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         endpoints = [object()]
         serializer = object()
 
+        class MagicMockIgnoreArgs(mock.MagicMock):
+            """MagicMock ignores arguments.
+
+            A MagicMock which can never misinterpret the arguments passed to
+            it during construction.
+            """
+
+            def __init__(self, *args, **kwargs):
+                super(MagicMockIgnoreArgs, self).__init__()
+
         server = oslo_messaging.get_rpc_server(transport, target, endpoints,
                                                serializer=serializer)
         # Mocking executor
-        server._executor = mock.Mock()
+        server._executor_cls = MagicMockIgnoreArgs
+        server.listener = MagicMockIgnoreArgs()
+        server.dispatcher = MagicMockIgnoreArgs()
         # Here assigning executor's listener object to listener variable
-        # before calling wait method, beacuse in wait method we are
+        # before calling wait method, because in wait method we are
         # setting executor to None.
-        listener = server._executor.listener
+        server.start()
+        listener = server.listener
+        server.stop()
         # call server wait method
         server.wait()
-        self.assertIsNone(server._executor)
         self.assertEqual(1, listener.cleanup.call_count)
 
-    @mock.patch('oslo_messaging._executors.impl_pooledexecutor.'
-                'PooledExecutor.wait')
-    def test_server_invalid_wait_running_server(self, mock_wait):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
-        target = oslo_messaging.Target(topic='foo', server='bar')
-        endpoints = [object()]
-        serializer = object()
-
-        server = oslo_messaging.get_rpc_server(transport, target, endpoints,
-                                               serializer=serializer,
-                                               executor='eventlet')
-        self.addCleanup(server.wait)
-        self.addCleanup(server.stop)
-        server.start()
-        with mock.patch('logging.Logger.warn') as warn:
-            server.wait()
-            warn.assert_called_with('wait() should be called after '
-                                    'stop() as it waits for existing '
-                                    'messages to finish processing')
-
-    @mock.patch('oslo_messaging._executors.impl_pooledexecutor.'
-                'PooledExecutor.wait')
-    def test_server_invalid_stop_from_other_thread(self, mock_wait):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
-        target = oslo_messaging.Target(topic='foo', server='bar')
-        endpoints = [object()]
-        serializer = object()
-
-        server = oslo_messaging.get_rpc_server(transport, target, endpoints,
-                                               serializer=serializer,
-                                               executor='eventlet')
-
-        t = test_utils.ServerThreadHelper(server)
-        t.start()
-        self.addCleanup(t.join)
-        self.addCleanup(t.stop)
-        with mock.patch('logging.Logger.warn') as warn:
-            server.stop()
-            warn.assert_called_with('start/stop/wait must be called '
-                                    'in the same thread')
-        with mock.patch('logging.Logger.warn') as warn:
-            server.wait()
-            warn.assert_called_with('start/stop/wait must be called '
-                                    'in the same thread')
-
     def test_no_target_server(self):
         transport = oslo_messaging.get_transport(self.conf, url='fake:')
 
@@ -296,7 +281,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
 
         self.assertIsNone(client.call({}, 'ping', arg=None))
         self.assertEqual(0, client.call({}, 'ping', arg=0))
-        self.assertEqual(False, client.call({}, 'ping', arg=False))
+        self.assertFalse(client.call({}, 'ping', arg=False))
         self.assertEqual([], client.call({}, 'ping', arg=[]))
         self.assertEqual({}, client.call({}, 'ping', arg={}))
         self.assertEqual('dsdsfoo', client.call({}, 'ping', arg='foo'))
@@ -316,7 +301,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         direct = client.prepare(server='testserver')
         self.assertIsNone(direct.call({}, 'ping', arg=None))
         self.assertEqual(0, client.call({}, 'ping', arg=0))
-        self.assertEqual(False, client.call({}, 'ping', arg=False))
+        self.assertFalse(client.call({}, 'ping', arg=False))
         self.assertEqual([], client.call({}, 'ping', arg=[]))
         self.assertEqual({}, client.call({}, 'ping', arg={}))
         self.assertEqual('dsdsfoo', direct.call({}, 'ping', arg='foo'))
@@ -471,9 +456,9 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
             return scenario
 
         for f in [filter_fanout_call, filter_same_topic_and_server]:
-            cls.scenarios = filter(f, cls.scenarios)
+            cls.scenarios = [i for i in cls.scenarios if f(i)]
         for m in [fanout_to_servers, single_topic_multi_endpoints]:
-            cls.scenarios = map(m, cls.scenarios)
+            cls.scenarios = [m(i) for i in cls.scenarios]
 
     def __init__(self, *args):
         super(TestMultipleServers, self).__init__(*args)
@@ -548,3 +533,307 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
 
 
 TestMultipleServers.generate_scenarios()
+
+
+class TestServerLocking(test_utils.BaseTestCase):
+    def setUp(self):
+        super(TestServerLocking, self).setUp(conf=cfg.ConfigOpts())
+
+        def _logmethod(name):
+            def method(self, *args, **kwargs):
+                with self._lock:
+                    self._calls.append(name)
+            return method
+
+        executors = []
+
+        class FakeExecutor(object):
+            def __init__(self, *args, **kwargs):
+                self._lock = threading.Lock()
+                self._calls = []
+                self.listener = mock.MagicMock()
+                executors.append(self)
+
+            submit = _logmethod('submit')
+            shutdown = _logmethod('shutdown')
+
+        self.executors = executors
+
+        self.server = oslo_messaging.MessageHandlingServer(mock.Mock(),
+                                                           mock.Mock())
+        self.server._executor_cls = FakeExecutor
+
+    def test_start_stop_wait(self):
+        # Test a simple execution of start, stop, wait in order
+
+        eventlet.spawn(self.server.start)
+        self.server.stop()
+        self.server.wait()
+
+        self.assertEqual(len(self.executors), 2)
+        self.assertEqual(self.executors[0]._calls, ['shutdown'])
+        self.assertEqual(self.executors[1]._calls, ['submit', 'shutdown'])
+        self.assertTrue(self.server.listener.cleanup.called)
+
+    def test_reversed_order(self):
+        # Test that if we call wait, stop, start, these will be correctly
+        # reordered
+
+        eventlet.spawn(self.server.wait)
+        # This is non-deterministic, but there's not a great deal we can do
+        # about that
+        eventlet.sleep(0)
+
+        eventlet.spawn(self.server.stop)
+        eventlet.sleep(0)
+
+        eventlet.spawn(self.server.start)
+
+        self.server.wait()
+
+        self.assertEqual(len(self.executors), 2)
+        self.assertEqual(self.executors[0]._calls, ['shutdown'])
+        self.assertEqual(self.executors[1]._calls, ['submit', 'shutdown'])
+
+    def test_wait_for_running_task(self):
+        # Test that if 2 threads call a method simultaneously, both will wait,
+        # but only 1 will call the underlying executor method.
+
+        start_event = threading.Event()
+        finish_event = threading.Event()
+
+        running_event = threading.Event()
+        done_event = threading.Event()
+
+        _runner = [None]
+
+        class SteppingFakeExecutor(self.server._executor_cls):
+            def __init__(self, *args, **kwargs):
+                # Tell the test which thread won the race
+                _runner[0] = eventlet.getcurrent()
+                running_event.set()
+
+                start_event.wait()
+                super(SteppingFakeExecutor, self).__init__(*args, **kwargs)
+                done_event.set()
+
+                finish_event.wait()
+
+        self.server._executor_cls = SteppingFakeExecutor
+
+        start1 = eventlet.spawn(self.server.start)
+        start2 = eventlet.spawn(self.server.start)
+
+        # Wait until one of the threads starts running
+        running_event.wait()
+        runner = _runner[0]
+        waiter = start2 if runner == start1 else start2
+
+        waiter_finished = threading.Event()
+        waiter.link(lambda _: waiter_finished.set())
+
+        # At this point, runner is running start(), and waiter() is waiting for
+        # it to complete. runner has not yet logged anything.
+        self.assertEqual(0, len(self.executors))
+        self.assertFalse(waiter_finished.is_set())
+
+        # Let the runner log the call
+        start_event.set()
+        done_event.wait()
+
+        # We haven't signalled completion yet, so submit shouldn't have run
+        self.assertEqual(1, len(self.executors))
+        self.assertEqual(self.executors[0]._calls, [])
+        self.assertFalse(waiter_finished.is_set())
+
+        # Let the runner complete
+        finish_event.set()
+        waiter.wait()
+        runner.wait()
+
+        # Check that both threads have finished, start was only called once,
+        # and execute ran
+        self.assertTrue(waiter_finished.is_set())
+        self.assertEqual(2, len(self.executors))
+        self.assertEqual(self.executors[0]._calls, [])
+        self.assertEqual(self.executors[1]._calls, ['submit'])
+
+    def test_start_stop_wait_stop_wait(self):
+        # Test that we behave correctly when calling stop/wait more than once.
+        # Subsequent calls should be noops.
+
+        self.server.start()
+        self.server.stop()
+        self.server.wait()
+        self.server.stop()
+        self.server.wait()
+
+        self.assertEqual(len(self.executors), 2)
+        self.assertEqual(self.executors[0]._calls, ['shutdown'])
+        self.assertEqual(self.executors[1]._calls, ['submit', 'shutdown'])
+        self.assertTrue(self.server.listener.cleanup.called)
+
+    def test_state_wrapping(self):
+        # Test that we behave correctly if a thread waits, and the server state
+        # has wrapped when it it next scheduled
+
+        # Ensure that if 2 threads wait for the completion of 'start', the
+        # first will wait until complete_event is signalled, but the second
+        # will continue
+        complete_event = threading.Event()
+        complete_waiting_callback = threading.Event()
+
+        start_state = self.server._states['start']
+        old_wait_for_completion = start_state.wait_for_completion
+        waited = [False]
+
+        def new_wait_for_completion(*args, **kwargs):
+            if not waited[0]:
+                waited[0] = True
+                complete_waiting_callback.set()
+                complete_event.wait()
+            old_wait_for_completion(*args, **kwargs)
+
+        start_state.wait_for_completion = new_wait_for_completion
+
+        # thread1 will wait for start to complete until we signal it
+        thread1 = eventlet.spawn(self.server.stop)
+        thread1_finished = threading.Event()
+        thread1.link(lambda _: thread1_finished.set())
+
+        self.server.start()
+        complete_waiting_callback.wait()
+
+        # The server should have started, but stop should not have been called
+        self.assertEqual(2, len(self.executors))
+        self.assertEqual(self.executors[0]._calls, [])
+        self.assertEqual(self.executors[1]._calls, ['submit'])
+        self.assertFalse(thread1_finished.is_set())
+
+        self.server.stop()
+        self.server.wait()
+
+        # We should have gone through all the states, and thread1 should still
+        # be waiting
+        self.assertEqual(2, len(self.executors))
+        self.assertEqual(self.executors[0]._calls, ['shutdown'])
+        self.assertEqual(self.executors[1]._calls, ['submit', 'shutdown'])
+        self.assertFalse(thread1_finished.is_set())
+
+        # Start again
+        self.server.start()
+
+        # We should now record 4 executors (2 for each server)
+        self.assertEqual(4, len(self.executors))
+        self.assertEqual(self.executors[0]._calls, ['shutdown'])
+        self.assertEqual(self.executors[1]._calls, ['submit', 'shutdown'])
+        self.assertEqual(self.executors[2]._calls, [])
+        self.assertEqual(self.executors[3]._calls, ['submit'])
+        self.assertFalse(thread1_finished.is_set())
+
+        # Allow thread1 to complete
+        complete_event.set()
+        thread1_finished.wait()
+
+        # thread1 should now have finished, and stop should not have been
+        # called again on either the first or second executor
+        self.assertEqual(4, len(self.executors))
+        self.assertEqual(self.executors[0]._calls, ['shutdown'])
+        self.assertEqual(self.executors[1]._calls, ['submit', 'shutdown'])
+        self.assertEqual(self.executors[2]._calls, [])
+        self.assertEqual(self.executors[3]._calls, ['submit'])
+        self.assertTrue(thread1_finished.is_set())
+
+    @mock.patch.object(server_module, 'DEFAULT_LOG_AFTER', 1)
+    @mock.patch.object(server_module, 'LOG')
+    def test_logging(self, mock_log):
+        # Test that we generate a log message if we wait longer than
+        # DEFAULT_LOG_AFTER
+
+        log_event = threading.Event()
+        mock_log.warning.side_effect = lambda _, __: log_event.set()
+
+        # Call stop without calling start. We should log a wait after 1 second
+        thread = eventlet.spawn(self.server.stop)
+        log_event.wait()
+
+        # Redundant given that we already waited, but it's nice to assert
+        self.assertTrue(mock_log.warning.called)
+        thread.kill()
+
+    @mock.patch.object(server_module, 'LOG')
+    def test_logging_explicit_wait(self, mock_log):
+        # Test that we generate a log message if we wait longer than
+        # the number of seconds passed to log_after
+
+        log_event = threading.Event()
+        mock_log.warning.side_effect = lambda _, __: log_event.set()
+
+        # Call stop without calling start. We should log a wait after 1 second
+        thread = eventlet.spawn(self.server.stop, log_after=1)
+        log_event.wait()
+
+        # Redundant given that we already waited, but it's nice to assert
+        self.assertTrue(mock_log.warning.called)
+        thread.kill()
+
+    @mock.patch.object(server_module, 'LOG')
+    def test_logging_with_timeout(self, mock_log):
+        # Test that we log a message after log_after seconds if we've also
+        # specified an absolute timeout
+
+        log_event = threading.Event()
+        mock_log.warning.side_effect = lambda _, __: log_event.set()
+
+        # Call stop without calling start. We should log a wait after 1 second
+        thread = eventlet.spawn(self.server.stop, log_after=1, timeout=2)
+        log_event.wait()
+
+        # Redundant given that we already waited, but it's nice to assert
+        self.assertTrue(mock_log.warning.called)
+        thread.kill()
+
+    def test_timeout_wait(self):
+        # Test that we will eventually timeout when passing the timeout option
+        # if a preceding condition is not satisfied.
+
+        self.assertRaises(server_module.TaskTimeout,
+                          self.server.stop, timeout=1)
+
+    def test_timeout_running(self):
+        # Test that we will eventually timeout if we're waiting for another
+        # thread to complete this task
+
+        # Start the server, which will also instantiate an executor
+        self.server.start()
+        self.server.stop()
+        shutdown_called = threading.Event()
+
+        # Patch the executor's stop method to be very slow
+        def slow_shutdown(wait):
+            shutdown_called.set()
+            eventlet.sleep(10)
+        self.executors[0].shutdown = slow_shutdown
+
+        # Call wait in a new thread
+        thread = eventlet.spawn(self.server.wait)
+
+        # Wait until the thread is in the slow stop method
+        shutdown_called.wait()
+
+        # Call wait again in the main thread with a timeout
+        self.assertRaises(server_module.TaskTimeout,
+                          self.server.wait, timeout=1)
+        thread.kill()
+
+    @mock.patch.object(server_module, 'LOG')
+    def test_log_after_zero(self, mock_log):
+        # Test that we do not log a message after DEFAULT_LOG_AFTER if the
+        # caller gave log_after=1
+
+        # Call stop without calling start.
+        self.assertRaises(server_module.TaskTimeout,
+                          self.server.stop, log_after=0, timeout=2)
+
+        # We timed out. Ensure we didn't log anything.
+        self.assertFalse(mock_log.warning.called)
